#!/usr/bin/env python
from __future__ import print_function

import sys
import argparse

import Queue
import time

import cv2
import numpy as np

import lmdb
import caffe

import dataHelper

psfScale = 10.0
maxME = 255
binCount = 1000

def computeMSE(net, images, targets):

    if net is None:
        return [0]*len(targets), []

    scaleConstant = 0.004

    shape = net.blobs['data'].data.shape

    data = np.zeros([len(images), 3, shape[2], shape[3]], dtype=np.float32)

    for i, img in enumerate(images):
        b = (img.shape[0] - shape[2]) / 2
        croppedImage = img[b:b+shape[2], b:b+shape[2], :].astype(np.float32)
        data[i, :, :, :] = croppedImage.transpose([2, 0, 1]).reshape(1, 3, shape[2], shape[2])

    data[:, 0, :, :] -= 103.939
    data[:, 1, :, :] -= 116.779
    data[:, 2, :, :] -= 123.68
    data *= scaleConstant

    out = net.forward_all(data=data)

    out = out['CONVX_12']
    out /= scaleConstant
    out[:, 0, :, :] += 103.939
    out[:, 1, :, :] += 116.779
    out[:, 2, :, :] += 123.68
    cnnOut = []

    mse = [0]*len(targets)
    for i, target in enumerate(targets):
        cnnOut.append(out[i, :, :, :].reshape(out.shape[1], out.shape[2], out.shape[3]).transpose(1, 2, 0))
        mse[i] = np.mean((cnnOut[-1] - target)**2)

        #print(str(mse[i]))
        #cv2.imshow("original", images[i])
        #cv2.imshow("out", out[i, :, :, :].reshape(3, out.shape[2], out.shape[3]).transpose(1, 2, 0)/255)
        #cv2.imshow("target", target)
        #cv2.waitKey()

    return mse, cnnOut

def totalVariance(images):
    return [np.mean(np.sqrt(
                np.square(cv2.Sobel(img, cv2.CV_64F, 0, 1, ksize=3))
              + np.square(cv2.Sobel(img, cv2.CV_64F, 1, 0, ksize=3) + 0.0000001)
            )) for img in images]

def updateHistogram(val, maxVal, histogram):
    binCount = int(len(histogram))
    discVal = np.minimum((val / maxVal) * (binCount - 1) + 0.5, binCount - 1).astype(np.int64)
    histogram += np.bincount(discVal, minlength=binCount)

def getProbabilities(histogram, val, maxVal):
    cum = np.cumsum(histogram)
    cum /= cum[-1]

    binCount = len(histogram)
    discVal = np.minimum((val / maxVal) * (binCount - 1) + 0.5, binCount - 1).astype(np.int64)

    return [cum[x] for x in discVal]

def writeData(env1, data1, env2, data2, env3, data3):
    with env1.begin(write=True) as txn1:
     with env2.begin(write=True) as txn2:
      with env3.begin(write=True) as txn3:
        txn1.cursor().putmulti(data1, overwrite=True)
        txn2.cursor().putmulti(data2, overwrite=True)
        txn3.cursor().putmulti(data3, overwrite=True)

def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--patch_size", help="Size of extracted blurred patches.", type=int, default=64)
    parser.add_argument("--border_size",
                        help="Border which will be added to patch_size and removed after geometric transformations.",
                        type=int, default=0)

    parser.add_argument("--rotationX_sdev",
                        help="Standard deviation of camera rotation around X-axis in degrees.",
                        type=float, default=0.0)
    parser.add_argument("--rotationY_sdev",
                        help="Standard deviation of camera rotation around Y-axis in degrees.",
                        type=float, default=0.0)
    parser.add_argument("--rotationZ_sdev",
                        help="Standard deviation of camera rotation around Z-axis in degrees.",
                        type=float, default=0.0)
    parser.add_argument("--min_page_fov",
                        help="Minimal field of view of the wole page in degrees.",
                        type=float, default=10.0)
    parser.add_argument("--max_page_fov",
                        help="Maximal field of view of the wole page in degrees.",
                        type=float, default=60.0)

    parser.add_argument("-f", "--scale_factor",
                        help="Optional image decimation factor.",
                        type=float, default=1.0)
    parser.add_argument("-n", "--max_noise",
                        help="Maximum sdev of additive gaussian noise (sdev sampled uniformly from [min_noise, max_noise]).",
                        type=float, default=10)
    parser.add_argument("--min_noise",
                        help="Minimum sdev of additive gaussian noise (sdev sampled uniformly from [min_noise, max_noise]).",
                        type=float, default=0)
    parser.add_argument("-p", "--max_dof_radius",
                        help="Maximum radius of disc PSF for depth of field simulation. Discs are sampled uniformly from [min_dof_radius, max_dof_radius].",
                        type=float, default=4.0)
    parser.add_argument("--min_dof_radius",
                        help="Minimum radius of disc PSF for depth of field simulation. Discs are sampled uniformly from [min_dof_radius, max_psf_radius].",
                        type=float, default=0)
    parser.add_argument("-m", "--max_motion_radius",
                        help="Maximum radius (in pixels) of motion blur PSF. Radii are sampled uniformly from [min_motion_radius, max_motion_radius].",
                        type=int, default=0)
    parser.add_argument("--min_motion_radius",
                        help="Minimum radius (in pixels) of motion blur PSF. Radii are sampled uniformly from [min_motion_radius, max_motion_radius].",
                        type=int, default=0)
    parser.add_argument("--motion_halflife",
                        help="Smoothness of the motion blur kernels. It is a half-life of running running average as a fraction of total blur time.",
                        type=float, default=0.5)


    parser.add_argument("--gamma_sdev",
                        help="Standard deviation of gamma correction. The mean is 1.",
                        type=float, default=0.0)
    parser.add_argument("--jpeg_min",
                        help="Minimum JPEG compression quality.",
                        type=int, default=100)
    parser.add_argument("--jpeg_max",
                        help="Maximum JPEG compression quality.",
                        type=int, default=100)
    parser.add_argument("--gray_min",
                        help="Minimum image value for gray-scale scaling.",
                        type=float, default=0)
    parser.add_argument("--gray_max",
                        help="Maximum image value for gray-scale scaling.",
                        type=int, default=255)
    parser.add_argument("--color_sdev",
                        help="Standard deviation of color channel multiplier (changing color balance).",
                        type=float, default=0.0)

    parser.add_argument("--rank_power",
                        help="Higher values give more uniform distribution of example difficulty. Lower values produce harder examples.",
                        type=float, default=1.0)

    parser.add_argument("-mb", dest="mb",
                        help="Simulate motion blur.",
                        action="store_true", default=False)
    parser.add_argument("-mbamin","--motion_blur_angle_min",
                        help="The minimum inclusive angle of a motion blur to the x axe in degrees. The Y axe heads up.",
                        type=float, default=90.)
    parser.add_argument("-mbamax","--motion_blur_angle_max",
                        help="The maximum exclusive angle of a motion blur to the x axe in degrees. The Y axe heads up.",
                        type=float, default=90.)
    parser.add_argument("-mblmin","--motion_blur_length_min",
                        help="The minimum inclusive length of a motion blur in subpixel - float.",
                        type=float, default=0.)
    parser.add_argument("-mblmax","--motion_blur_length_max",
                        help="The maximum exclusive length of a motion blur in subpixel - float.",
                        type=float, default=11.)

    parser.add_argument( "-o", "--blurred_output", help="Blurred patch output LMDB.",)
    parser.add_argument( "-O", "--original_output", help="Original patch output LMDB.")
    parser.add_argument( "-P", "--psf_output", help="PSF output LMDB.")
    parser.add_argument( "-I", "--write_images", help="Write png images into working directory.", action='store_true')

    parser.add_argument( "-d", "--image_dir", help="Optional path to the images.", default="")

    parser.add_argument("--model_def", help="Model definition file.", default='')
    parser.add_argument("--pretrained_model", help="Trained model weights file.", default='')
    parser.add_argument("--gpu", action='store_true', help="Switch for gpu computation.")

    parser.add_argument("--start_id", help="Initial ID for writing in databases.", type=int, default=0)
    parser.add_argument("--net_resolution", help="Resolution of CNN output.", type=int, required=True)
    parser.add_argument("--per_image", help="Patches extracted per image.", type=int, required=True)
    parser.add_argument("--max_key", help="Maximum key value. Key cycles within this range.", type=int, default=1000000000)

    args = parser.parse_args()
    patchSize = args.patch_size
    scaleFactor = args.scale_factor
    maxNoise = args.max_noise
    maxPSFRadius = args.max_dof_radius
    imageDir = args.image_dir
    perImage = args.per_image
    netRes = args.net_resolution
    doMotionBlur = args.mb
    vec_motionBlurLength = np.asarray([args.motion_blur_length_min, args.motion_blur_length_max])
    vec_motionBlurAngle =  np.asarray([args.motion_blur_angle_min, args.motion_blur_angle_max])
    maxKey = args.max_key
    startID = args.start_id

    RNG = np.random.RandomState()

    if imageDir:
        imageDir = imageDir + "/"

    net = None
    if args.pretrained_model:
        if hasattr(caffe, 'set_mode_gpu'):
            if args.gpu:
                print('GPU mode', file=sys.stderr)
                caffe.set_mode_gpu()
            net = caffe.Net(args.model_def, args.pretrained_model, caffe.TEST)
        else:
            if args.gpu:
                print('GPU mode', file=sys.stderr)
            net = caffe.Net(args.model_def, args.pretrained_model, gpu=args.gpu)

    print("patchSize %d, scaleFactor %f, maxNoise %f, maxPSFRadius %f"
          % (patchSize, scaleFactor, maxNoise, maxPSFRadius), file=sys.stderr)
    print("motionBlurLengthMin %f, motionBlurLengthMax %f, motionBlurAngleMin %f, motionBlurAngleMax %f"
          % (vec_motionBlurLength[0], vec_motionBlurLength[1], vec_motionBlurAngle[0], vec_motionBlurAngle[1]), file=sys.stderr)

    fileList = [x.strip() for x in sys.stdin]

    print("perImage %d" % perImage, file=sys.stderr)

    envOrig = lmdb.open(args.original_output, map_size=150000000000)
    envBlur = lmdb.open(args.blurred_output, map_size=150000000000)
    envPSF  = lmdb.open(args.psf_output, map_size=150000000000)

    v0x = 0.0
    v0y = 0.0

    histogram = np.zeros(binCount, dtype=np.float64)


    cOrig = []
    cBlur = []
    cPSF =  []

    imageQueue = Queue.Queue(32)
    reader = dataHelper.ReaderThread(imageQueue=imageQueue, fileList=fileList, imageDir=imageDir, scaleFactor=scaleFactor)
    reader.start()

    startTime = time.time()

    counter = 0
    while True:
        imOrg = imageQueue.get(block=True)
        if imOrg is None:
            break
        if len(imOrg.shape) != 3 and imOrg.shape[2] != 3:
            continue
        print(counter)
        if imOrg.shape[0] < 1.2*(patchSize + 2*args.border_size) or imOrg.shape[1] < 1.2*(patchSize + 2*args.border_size):
            print( "Image too small", imOrg.shape)
            continue

        minFOV = args.min_page_fov / max(imOrg.shape) * (patchSize + 2*args.border_size)
        maxFOV = args.max_page_fov / max(imOrg.shape) * (patchSize + 2*args.border_size)

        cropOrig = []
        cropBlur = []
        cropPsf= []
        cropParam = []

        for it in range(perImage):

            #get random crop
            cropOrigNew = dataHelper.generateRandomCrop(RNG, imOrg, patchSize + 2*args.border_size)
            if cropOrigNew is None:
                continue

            cropBlurNew = None
            cropPsfNew = None
            cropParamNew = None

            #geometric transformations
            rx = RNG.standard_normal() * args.rotationX_sdev
            ry = RNG.standard_normal() * args.rotationY_sdev
            rz = RNG.standard_normal() * args.rotationZ_sdev
            fov = RNG.uniform(minFOV, maxFOV)
            if rx != 0 or ry != 0 or rz != 0:
                cropOrigNew = dataHelper.warpPerspective(rx, ry, rz, fov, cropOrigNew)

            # viewing transformations PSF, color, compression, noise, ...
            psfRadius = RNG.uniform(args.min_dof_radius, maxPSFRadius)
            psf, center = dataHelper.generatePSF(psfRadius)

            motionRadius = int(RNG.uniform(args.min_motion_radius, args.max_motion_radius) + 0.5)
            motionPSF = dataHelper.generateMotionBlur(RNG, resolution=motionRadius*2+1, halflife=args.motion_halflife) # can't skip to keep correct RNG state
            if motionRadius > 0:
                border = motionRadius + 1
                psf = cv2.copyMakeBorder(psf, border, border, border, border, cv2.BORDER_CONSTANT, value=0)
                psf = cv2.filter2D(psf, -1, motionPSF, borderType=cv2.BORDER_REPLICATE)

            if doMotionBlur:
                # Lets do Motion blur
                mbPSF, slope, length = dataHelper.generateMotionBlurPSF(RNG, slope_deg=vec_motionBlurAngle, length=vec_motionBlurLength)
                print("Slope: %f, Length: %f, kernelSize: %f" % (slope, length, mbPSF.shape[1]))
                psf = dataHelper.stackKernels(mbPSF, psf)
            # Is the normalization here necessary? The mbPS has been already normalized.
            psf = psf / np.sum(psf)
            cropPsfNew = psf

            cropBlurNew = cv2.filter2D(cropOrigNew, cv2.CV_32F, psf, borderType=cv2.BORDER_REPLICATE)

            # remove border
            if args.border_size != 0:
                b = args.border_size
                cropOrigNew = cropOrigNew[b:-b,b:-b,:]
                cropBlurNew = cropBlurNew[b:-b,b:-b,:]

            noiseSdev = RNG.uniform(args.min_noise, maxNoise)
            cropBlurNew = dataHelper.addAdditiveWhiteNoise(RNG, cropBlurNew, noiseSdev)

            colorCoeff = [2**(RNG.standard_normal()*args.color_sdev) for x in range(3)]
            if colorCoeff[0] != 1 and colorCoeff[1] != 1 and colorCoeff[2] != 1:
                cropBlurNew = dataHelper.colorBalance(cropBlurNew, colorCoeff)

            minVal = RNG.uniform(0, args.gray_min)
            maxVal = RNG.uniform(args.gray_max, 255)
            if minVal != 0 and maxVal != 255:
                cropBlurNew = dataHelper.grayShiftScale(cropBlurNew, minVal, maxVal)

            gamma = 2**(RNG.standard_normal()*args.gamma_sdev)
            if gamma != 1:
                cropBlurNew = dataHelper.gammaCorrection(cropBlurNew, gamma, 255)

            jpegQuality = int(RNG.uniform(args.jpeg_min, args.jpeg_max))
            if jpegQuality != 100:
                cropBlurNew = dataHelper.addArtefactsJPEG(cropBlurNew, jpegQuality)

            np.clip(cropBlurNew, 0, 255, out=cropBlurNew)
            cropBlurNew += 0.5
            cropBlurNew = cropBlurNew.astype(np.uint8)

            #crop original
            border = int((cropOrigNew.shape[0] - netRes) / 2)
            if 2*border + netRes != cropOrigNew.shape[0]:
                print("Warning border is not wrong.", file=sys.stderr)

            cropOrigNew = cropOrigNew[border:border+netRes, border:border+netRes, :]

            cropParamNew = [psfRadius, motionRadius, noiseSdev, colorCoeff, minVal, maxVal, gamma, jpegQuality]

            if cropOrigNew is not None \
                and cropBlurNew is not None \
                and cropPsfNew is not None \
                and cropParamNew is not None \
                and cropOrigNew.shape[0] > 0:
                cropOrig.append(cropOrigNew)
                cropBlur.append(cropBlurNew)
                cropPsf.append(cropPsfNew)
                cropParam.append(cropParamNew)


        if net is None:
            ME = np.asarray(totalVariance(cropOrig))
        else:
            ME, cnnOut = computeMSE(net, cropBlur, cropOrig)
            ME = np.asarray(ME)**0.5

        updateHistogram(ME, maxME, histogram)
        prob = getProbabilities(histogram, ME, maxME)
        randVal = RNG.random_sample(len(prob))**args.rank_power

        cropBlur = [x[0] for x in zip(cropBlur, prob, randVal) if x[2] < x[1]]
        cropOrig = [x[0] for x in zip(cropOrig, prob, randVal) if x[2] < x[1]]
        cropPsf = [x[0] for x in zip(cropPsf, prob, randVal) if x[2] < x[1]]
        cropParam = [x[0] for x in zip(cropParam, prob, randVal) if x[2] < x[1]]
        #cnnOut = [x[0] for x in zip(cnnOut, prob, randVal) if x[2] < x[1]]

        print('Kept: %d' % len(cropBlur))

        for blur, orig, psf, param in zip(cropBlur, cropOrig, cropPsf, cropParam):

            if args.write_images:
                nameOrig = "%07d_orig.png" % (counter)
                nameBlur = "%07d_blur.png" % (counter)
                namePsf = "%07d_psf.png" % (counter)
                cv2.imwrite(nameOrig, orig)
                cv2.imwrite(nameBlur, blur)
                cv2.imwrite(namePsf, psf / np.max(psf) * 255)

            blur = blur.transpose((2, 0, 1)).astype(np.uint8, copy=False)
            orig = orig.transpose((2, 0, 1)).astype(np.uint8, copy=False)
            psf = psf.reshape((psf.shape[0], psf.shape[1], 1)).transpose((2, 0, 1)).astype(float)

            datumOrig = caffe.io.array_to_datum(orig)
            datumBlur = caffe.io.array_to_datum(blur)
            datumPSF  = caffe.io.array_to_datum(psf)

            key = "%09d" % ((counter % maxKey)+startID)

            cOrig.append((key, datumOrig.SerializeToString()))
            cBlur.append((key, datumBlur.SerializeToString()))
            cPSF.append( (key, datumPSF.SerializeToString()))

            print(' '.join([str(x) for x in param]))

            counter += 1
            if counter % 200 == 0:
                print("DONE %d in %f s-- Last key %s " % (counter, time.time() - startTime, key), file=sys.stderr)
                writeData(envOrig, cOrig, envBlur, cBlur, envPSF, cPSF)
                cOrig = []
                cBlur = []
                cPSF  = []

    writeData(envOrig, cOrig, envBlur, cBlur, envPSF, cPSF)


if __name__ == '__main__':
    main(sys.argv)
